#include "confidence_measures.h"
#include "time.h"

/***************************************************************************************************/
/*
* Confidence Measures
*
/***************************************************************************************************/

/***************************************************************************************************/
//	3.1 Minimum cost and local properties of the cost curve  
/***************************************************************************************************/
void matching_score_measure(InputArray c1, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1 = c1.getMat(),
		_confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		float* c1_ptr = _c1.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			float c_1 = c1_ptr[col];

		    confidence_map_ptr[col] = - c_1;
		}
	}
}

void maximum_margin(InputArray c1, InputArray c2m, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1 = c1.getMat(),
	    _c2m = c2m.getMat(), 
	    _confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		float* c1_ptr = _c1.ptr<float>(row);
		float* c2m_ptr = _c2m.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			float c_1 = c1_ptr[col];
			float c_2m = c2m_ptr[col];

			confidence_map_ptr[col] = c_2m - c_1;
		}
	}
}

void non_linear_margin(InputArray c1, InputArray c2m, float sigma, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1 = c1.getMat(),
		_c2m = c2m.getMat(),
		_confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		float* c1_ptr = _c1.ptr<float>(row);
		float* c2m_ptr = _c2m.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			float c_1 = c1_ptr[col];
			float c_2m = c2m_ptr[col];

			confidence_map_ptr[col] = exp(c_2m - c_1) / (2 * (sigma * sigma));
		}
	}
}


void curvature(InputArrayOfArrays costs, InputArray c1_idx, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
		_confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);

	int d1;
	float minimum, add1, add2;

	for (int row = 0; row < height; row++)
	{
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);
		
		for (int col = 0; col < width; col++)
		{
			d1 = c1_idx_ptr[col]; 
			minimum = _costs[d1].ptr<float>(row)[col];

			if (d1 - 1 < 0)
			{
				add2 = _costs[d1 + 1].ptr<float>(row)[col];
				add1 = _costs[d1 + 1].ptr<float>(row)[col];
			}
			else if (d1 + 1 > _costs.size() - 1)
			{
				add1 = _costs[d1 - 1].ptr<float>(row)[col];
				add2 = _costs[d1 - 1].ptr<float>(row)[col];
			}
			else
			{
				add1 = _costs[d1 - 1].ptr<float>(row)[col];
				add2 = _costs[d1 + 1].ptr<float>(row)[col];
			}

			confidence_map_ptr[col] = (-2 * minimum) + add1 + add2;		
		}
	}
}

void local_curve(InputArrayOfArrays costs, InputArray c1_idx, float gamma, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
		_confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);

	int d1;
	float minimum, add1, add2, max;

	for (int row = 0; row < height; row++)
	{
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);

		for (int col = 0; col < width; col++)
		{
			d1 = c1_idx_ptr[col];
			minimum = _costs[d1].ptr<float>(row)[col];

			if (d1 - 1 < 0)
			{
				add2 = _costs[d1 + 1].ptr<float>(row)[col];
				add1 = _costs[d1 + 1].ptr<float>(row)[col];
			}
			else if (d1 + 1 > _costs.size() - 1)
			{
				add1 = _costs[d1 - 1].ptr<float>(row)[col];
				add2 = _costs[d1 - 1].ptr<float>(row)[col];
			}
			else
			{
				add1 = _costs[d1 - 1].ptr<float>(row)[col];
				add2 = _costs[d1 + 1].ptr<float>(row)[col];
			}

			if (add1 >= add2) max = add1;
			else max = add2;

			confidence_map_ptr[col] = (max - minimum) / gamma;
		}
	}
}

void peak_ratio(InputArray c1, InputArray c2m, float epsilon, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1 = c1.getMat(), 
	    _c2m = c2m.getMat(),
		_confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		float* c1_ptr = _c1.ptr<float>(row);
		float* c2m_ptr = _c2m.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			float c_1 = c1_ptr[col];
			float c_2m = c2m_ptr[col];

			confidence_map_ptr[col] = (c_2m + epsilon) / (c_1 + epsilon); 	
		}
	}
}

void average_peak_ratio(InputArrayOfArrays costs, InputArray c1_idx, InputArray c2m_idx, float epsilon, int r, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
	    _c2m_idx = c2m_idx.getMat(),
	    _confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);

	for (int row = 0; row < height; row++)
	{
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);
		uchar* c2m_idx_ptr = _c2m_idx.ptr<uchar>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			uchar c_1_idx = c1_idx_ptr[col];
			uchar c_2m_idx = c2m_idx_ptr[col];
			float sum = 0;
			
			for (int i = -r; i <= r; i++)
			{
				for (int j = -r; j <= r; j++)
				{
					//c_2m_idx >= 0 if second local minimum doesn't exist
					if (row + i >= 0 && row + i < height && col + j >= 0 && col + j < width && c_2m_idx >= 0) 
					{
						//q is the running pixel 
						float q_c1 = _costs[c_1_idx].ptr<float>(row + i)[col + j];
						float q_c2m = _costs[c_2m_idx].ptr<float>(row + i)[col + j];

						sum += (q_c2m + epsilon) / (q_c1 + epsilon);
					}
				}
			}

			confidence_map_ptr[col] = sum / (2 * r + 1) * (2 * r + 1); //average
		}
	}
}

void weighted_peak_ratio(InputArrayOfArrays costs, InputArray left, InputArray c1_idx, InputArray c2m_idx, float epsilon, float threshold_color, int r, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _left,
		_c1_idx = c1_idx.getMat(),
		_c2m_idx = c2m_idx.getMat(),
		_confidence_map = confidence_map.getMat();

	left.getMat().convertTo(_left, CV_32F);

	Mat bgr[3];
	bool RGB = false;

	if (_left.channels() == 3)
	{
		RGB = true;
		split(_left, bgr);//split source 
	}
	
	vector<Mat> _costs; costs.getMatVector(_costs);

	for (int row = 0; row < height; row++)
	{
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);
		uchar* c2m_idx_ptr = _c2m_idx.ptr<uchar>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			uchar c_1_idx = c1_idx_ptr[col];
			uchar c_2m_idx = c2m_idx_ptr[col];
			float sum = 0;

			for (int i = -r; i <= r; i++)
			{
				for (int j = -r; j <= r; j++)
				{
					//c_2m_idx >= 0 if second local minimum doesn't exist
					if (row + i >= 0 && row + i < height && col + j >= 0 && col + j < width && c_2m_idx >= 0)
					{
						//q is the running pixel 
						float q_c1 = _costs[c_1_idx].ptr<float>(row + i)[col + j];
						float q_c2m = _costs[c_2m_idx].ptr<float>(row + i)[col + j];

						//compute distance color
						float distance = (RGB) ? sqrt(pow(bgr[0].ptr<float>(row)[col] - bgr[0].ptr<float>(row + i)[col + j], 2) +
													  pow(bgr[1].ptr<float>(row)[col] - bgr[1].ptr<float>(row + i)[col + j], 2) +
							                          pow(bgr[2].ptr<float>(row)[col] - bgr[2].ptr<float>(row + i)[col + j], 2)) 
							                   : fabs(_left.ptr<float>(row)[col] - _left.ptr<float>(row + i)[col + j]);

						if (distance < threshold_color)
							sum += (q_c2m + epsilon) / (q_c1 + epsilon);
					}
				}
			}

			confidence_map_ptr[col] = sum / (2 * r + 1) * (2 * r + 1); //average
		}
	}
}

void disparity_ambiguity_measure(InputArray c1_idx, InputArray c2_idx, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
	    _c2_idx = c2_idx.getMat(),
        _confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);
		uchar* c2_idx_ptr = _c2_idx.ptr<uchar>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			uchar d_1 = c1_idx_ptr[col]; 
			uchar d_2 = c2_idx_ptr[col];

			confidence_map_ptr[col] = - (float)fabs(d_1 - d_2);
		}
	}
}

void semi_global_energy(InputArray c1, InputArray disparity_L2R, int p1, int p2, int r, int height, int width, OutputArray confidence_map)
{
	//only 8 directions

	confidence_map.create(height, width, CV_32F);

	int win = 2 * r + 1;

	Mat _disparity_L2R = disparity_L2R.getMat(),
		_c1 = c1.getMat(),
		_confidence_map = confidence_map.getMat(),
		patch = Mat(win, win, CV_32F, Scalar(0));

	for (int row = 0; row < height; row++)
	{
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			float total_sum = _c1.ptr<float>(row)[col];

			int X = row, Y = col;
			int previous_X = X, previous_Y = Y;

			for (int direction = 0; direction < 8; direction++)
			{
				float partial_sum = 0;

				for (int k = 0; k < r; k++)
				{
					get_next_xy(&X, &Y, direction);

					if (X >= 0 && X < height && Y >= 0 && Y< width)
					{
						float c_q = _c1.ptr<float>(X)[Y];
						float d_p = _disparity_L2R.ptr<float>(row)[col];
						float d_q = _disparity_L2R.ptr<float>(X)[Y];

						if (d_p >= 0 && d_q >= 0)
						{
							if (abs(d_q - d_p) == 1)
							{
								partial_sum += c_q + p1;
							}
							else if (abs(d_q - d_p) > 1)
							{
								partial_sum += c_q + p2;
							}
							else
							{
								partial_sum += c_q;
							}
						}
					}
				}

				total_sum += partial_sum;

				X = previous_X;
				Y = previous_Y;
			}

			confidence_map_ptr[col] = - total_sum;
		}
	}
}

/***************************************************************************************************/
//	3.2 Analysis of the entire cost curve
/***************************************************************************************************/
void perturbation_measure(InputArrayOfArrays costs, InputArray c1, InputArray c1_idx, float s, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(), 
		_c1 = c1.getMat(),
		_confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);

	for (int row = 0; row < height; row++)
	{
		float* c1_ptr = _c1.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);

		for (int col = 0; col < width; col++)
		{
			float sum = 0;
			float c_1 = c1_ptr[col];
			uchar d_1 = c1_idx_ptr[col];
			
			for (int d = 0; d < _costs.size(); d++)
			{
				float c_d = _costs[d].ptr<float>(row)[col];
				float diff = (c_1 - c_d);

				if (d != d_1)
					sum += exp(- diff * diff) / (s * s);
			}

			_confidence_map.ptr<float>(row)[col] = - sum;
		}
	}
}

void maximum_likelihood_measure(InputArrayOfArrays costs, InputArray c1, float sigma, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _sum, 
		_c1 = c1.getMat(),
		_confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);

	for (int row = 0; row < height; row++)
	{
		for (int col = 0; col < width; col++)
		{
			float den = 0;

			for (int d = 0; d < _costs.size(); d++)
			{
				float value = _costs[d].ptr<float>(row)[col];

				den += exp(-value / (2 * sigma * sigma));
			}

			float c_1 = _c1.ptr<float>(row)[col];

			if (den != 0)
				_confidence_map.ptr<float>(row)[col] = exp(-c_1 / 2 * sigma * sigma) / den;
			else
				_confidence_map.ptr<float>(row)[col] = 0; 
		}
	}
}

void attainable_maximum_likelihood(InputArrayOfArrays costs, InputArray c1, float sigma, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1 = c1.getMat(),
		_confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);

	for (int row = 0; row < height; row++)
	{
		for (int col = 0; col < width; col++)
		{
			float c_1 = _c1.ptr<float>(row)[col];
			float den = 0;

			for (int d = 0; d < _costs.size(); d++)
			{
				float value = _costs[d].ptr<float>(row)[col];

				den += exp(-(pow(value - c_1,2)) / (2 * sigma * sigma));
			}

			if (den != 0)
				_confidence_map.ptr<float>(row)[col] = 1 / den;
			else
				_confidence_map.ptr<float>(row)[col] = 0;
		}
	}
}

void local_minima_in_neighborhood(InputArrayOfArrays local_minima, InputArray c1_idx, int r, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
		_confidence_map = confidence_map.getMat();

	vector<Mat> _local_minima; local_minima.getMatVector(_local_minima);

	for (int row = 0; row < height; row++)
	{
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			uchar c_1_idx = c1_idx_ptr[col];
			float sum = 0;

			for (int i = -r; i <= r; i++)
			{
				for (int j = -r; j <= r; j++)
				{
					if (row + i >= 0 && row + i < height && col + j >= 0 && col + j < width)
					{
						sum += _local_minima[c_1_idx].ptr<uchar>(row + i)[col + j];
					}
				}
			}

			confidence_map_ptr[col] = sum;
		}
	}
}

void winner_margin(InputArray c1, InputArray c2m, InputArray c_sum, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1 = c1.getMat(), 
	    _c2m = c2m.getMat(), 
	    _c_sum = c2m.getMat(), 
    	_confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		float* c1_ptr = _c1.ptr<float>(row);
		float* c2m_ptr = _c2m.ptr<float>(row);
		float* c_sum_ptr = _c_sum.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			float c_1 = c1_ptr[col];
			float c_2m = c2m_ptr[col];
			float c_sum = c_sum_ptr[col];

			if (c_sum > 0)confidence_map_ptr[col] = (c_2m - c_1) / c_sum;
			else confidence_map_ptr[col] = 0;
		}
	}
}

void negative_entropy_measure(InputArrayOfArrays costs, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);

	for (int row = 0; row < height; row++)
	{
		for (int col = 0; col < width; col++)
		{
			float negative_entropy = 0;
			float den = 0;

			for (int d = 0; d < _costs.size(); d++)
			{
				float value = _costs[d].ptr<float>(row)[col];

				den += exp(-value); 
			}
			
			for (int d = 0; d < _costs.size(); d++)
			{
				float value = _costs[d].ptr<float>(row)[col];

				float p_d = exp(-value) / den;

				negative_entropy += (p_d * log2(p_d)); 	
			}

			_confidence_map.ptr<float>(row)[col] = negative_entropy;
		}
	}
}

/***************************************************************************************************/
//	3.3 Left and right consistency
/***************************************************************************************************/
void left_right_consistency_check(InputArray disparity_L2R, InputArray disparity_R2L, int disp_scale, int height, int width, OutputArray confidence_map){

	confidence_map.create(height, width, CV_32F);

	Mat _left = disparity_L2R.getMat(),
	    _right = disparity_R2L.getMat(),
	    _confidence_map = confidence_map.getMat();

	_confidence_map.setTo(Scalar(0));

	for (int row = 0; row < height; row++)
	{
		float* left_ptr = _left.ptr<float>(row);
		float* right_ptr = _right.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			if (left_ptr[col] >= 0)
			{
				int j_right = col - (left_ptr[col]) / disp_scale;

				if (j_right < 0)
					confidence_map_ptr[col] = 0;
				else
				{
					if (right_ptr[j_right] >= 0)
					{
						int jRefp = j_right + (right_ptr[j_right]) / disp_scale;

						confidence_map_ptr[col] = -(float)fabs(col - jRefp);
					}
				}
			}
		}
	}
}

void left_right_difference(InputArray c1_L, InputArray c2_L, InputArray c1_R, InputArray disparity_L, int disp_scale, float epsilon, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_L = c1_L.getMat(),
		_c2_L = c2_L.getMat(), 
	    _c1_R = c1_R.getMat(), 
	    _disparity_L = disparity_L.getMat(),
        _confidence_map = confidence_map.getMat();

	_confidence_map.setTo(Scalar(0));

	for (int row = 0; row < height; row++)
	{
		float* c1_L_ptr = _c1_L.ptr<float>(row);
		float* c2_L_ptr = _c2_L.ptr<float>(row);
		float* disparity_L_ptr = _disparity_L.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			float c1_l = c1_L_ptr[col]; 
			float c2_l = c2_L_ptr[col];
			int d1_l = (int)disparity_L_ptr[col]/disp_scale;
			
			if (col - d1_l >= 0 && d1_l>= 0)
			{
				float den = (float)fabs(c1_l - _c1_R.ptr<float>(row)[col - d1_l]); 

				confidence_map_ptr[col] = ((c2_l - c1_l) + epsilon) / (den + epsilon);
			}		
		}
	}
}

void zero_mean_sum_of_absolute_differences(InputArray left, InputArray right, InputArray c1, InputArray c1_idx, int r, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _left, _right,
	    _c1 = c1.getMat(), _c1_idx = c1_idx.getMat(),
		_abs_diff = Mat(left.size(), CV_32F),
		_abs_box_diff = Mat(left.size(), CV_32F),
		_mean_L, _mean_R,
		_confidence_map = confidence_map.getMat();

	left.getMat().convertTo(_left, CV_32F);
	right.getMat().convertTo(_right, CV_32F);

	_mean_L.create(left.size(), CV_32F);
	_mean_R.create(right.size(), CV_32F);

	//compute absolute difference (point operator)
	for (int row = 0; row < _left.rows; row++)
	{
		float* left_ptr = _left.ptr<float>(row);
		float* right_ptr = _right.ptr<float>(row);
		float* abs_diff_ptr = _abs_diff.ptr<float>(row);
		uchar* index_ptr = _c1_idx.ptr<uchar>(row);

		for (int col = 0; col < _left.cols; col++)
		{
			uchar index = index_ptr[col];
			int col_R = (col - index < 0) ? 0 : col - index;
			abs_diff_ptr[col] = fabs(left_ptr[col] - right_ptr[col_R]);
		}
	}

	box_filter(_left, r, _mean_L);
	box_filter(_right, r, _mean_R);
	box_filter(_abs_diff, r, _abs_box_diff);
	
	_mean_L = _mean_L / ((2 * r + 1) * (2 * r + 1));
	_mean_R = _mean_R / ((2 * r + 1) * (2 * r + 1));

	for (int row = 0; row < _left.rows; row++)
	{
		float* mean_L_ptr = _mean_L.ptr<float>(row);
		float* mean_R_ptr = _mean_R.ptr<float>(row);
		float* abs_box_diff_ptr = _abs_box_diff.ptr<float>(row);
		float* left_ptr = _left.ptr<float>(row);
		float* right_ptr = _right.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		uchar* index_ptr = _c1_idx.ptr<uchar>(row);

		for (int col = 0; col < _left.cols; col++)
		{
			uchar index = index_ptr[col];
			int col_R = (col - index < 0) ? 0 : col - index;
			confidence_map_ptr[col] = - fabs(abs_box_diff_ptr[col] - mean_L_ptr[col] + mean_R_ptr[col_R]);
		}
	}
}

void uniqueness_constraint(InputArray c1, InputArray c1_idx, int height, int width, OutputArray confidence_map_1, OutputArray confidence_map_2, OutputArray confidence_map_3)
{
	confidence_map_1.create(height, width, CV_32F);
	confidence_map_2.create(height, width, CV_32F);
	confidence_map_3.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
		_c1 = c1.getMat(),
		_confidence_map_1 = confidence_map_1.getMat(),
		_confidence_map_2 = confidence_map_2.getMat(),
		_confidence_map_3 = confidence_map_3.getMat(),
		_occurrence = Mat(height, width, CV_32F),
		_cost_register = Mat(1, width, CV_32F),
		_index_register = Mat(1, width, CV_32S);

	_confidence_map_1.setTo(Scalar(0));
	_occurrence.setTo(Scalar(0));
	_confidence_map_3.setTo(Scalar(-1));

	for (int row = 0; row < height; row++)
	{
		_cost_register.setTo(Scalar(-1));
		_index_register.setTo(Scalar(-1));

		float* c1_ptr = _c1.ptr<float>(row);
		float* cost_register_ptr = _cost_register.ptr<float>(0);
		int* index_register_ptr = _index_register.ptr<int>(0);
		float* confidence_map_1_ptr = _confidence_map_1.ptr<float>(row);
		float* occurrence_ptr = _occurrence.ptr<float>(row);
		float* confidence_map_3_ptr = _confidence_map_3.ptr<float>(row);
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);
		
		for (int col = 0; col < width; col++)
		{
			int index_r = col - c1_idx_ptr[col]; 

			if (index_r >= 0)
			{
				float cost = c1_ptr[col]; 
				float best_cost = cost_register_ptr[index_r]; 

				if (best_cost >= 0)
				{
					occurrence_ptr[index_r]++;

					if (cost < best_cost) 
					{
						cost_register_ptr[index_r] = cost; 
						int old_index = index_register_ptr[index_r]; 
						confidence_map_1_ptr[old_index] = 0;
						confidence_map_1_ptr[col] = 1;
						confidence_map_3_ptr[col] = - cost;
						index_register_ptr[index_r] = col; 
					}
					else
						confidence_map_1_ptr[col] = 0;		
				}
				else
				{
					cost_register_ptr[index_r] = cost; 
					index_register_ptr[index_r] = col; 
					confidence_map_1_ptr[col] = 1;
					occurrence_ptr[index_r]++;
					confidence_map_3_ptr[col] = - cost;
				}
			}
		}
	}

	for (int row = 0; row < height; row++)
	{
		float* occurrence_ptr = _occurrence.ptr<float>(row);
		float* confidence_map_2_ptr = _confidence_map_2.ptr<float>(row);
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);

		for (int col = 0; col < width; col++)
		{
			int index_r = col - c1_idx_ptr[col];
			confidence_map_2_ptr[col] = - occurrence_ptr[index_r];
		}
	}
}

void asymmetric_consistency_check(InputArray c1, InputArray c1_idx, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
		_c1 = c1.getMat(),
		_confidence_map = confidence_map.getMat(),
		_cost_register = Mat(1, width, CV_32F),
		_disparity_register = Mat(1, width, CV_32S),
		_index_register = Mat(1, width, CV_32S);

	_confidence_map.setTo(Scalar(0));

	for (int row = 0; row < height; row++)
	{
		_cost_register.setTo(Scalar(-1));
		_disparity_register.setTo(Scalar(-1));
		_index_register.setTo(Scalar(-1));

		float* c1_ptr = _c1.ptr<float>(row);
		float* cost_register_ptr = _cost_register.ptr<float>(0);
		int* disparity_register_ptr = _disparity_register.ptr<int>(0);
		int* index_register_ptr = _index_register.ptr<int>(0);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);
		
		for (int col = 0; col < width; col++)
		{
			int index_r = col - c1_idx_ptr[col];
			int disp = c1_idx_ptr[col];

			if (index_r >= 0)
			{
				float cost = c1_ptr[col];
				float best_cost = cost_register_ptr[index_r];
				int max_disp = disparity_register_ptr[index_r];

				if (max_disp >= 0)
				{
					if (disp > max_disp)
					{
						disparity_register_ptr[index_r] = disp;
						int old_index = index_register_ptr[index_r];
						confidence_map_ptr[old_index] = 0;
						confidence_map_ptr[col] = 1;
						index_register_ptr[index_r] = col;

						if (cost >= best_cost)
						{
							confidence_map_ptr[col] = 0;
						}
						else
						{
							cost_register_ptr[index_r] = cost;
						}
					}
				}
				else
				{
					cost_register_ptr[index_r] = cost;
					disparity_register_ptr[index_r] = disp;
					index_register_ptr[index_r] = col;
					confidence_map_ptr[col] = 1;
				}
			}
		}
	}
}

/***************************************************************************************************/
//	3.4 Disparity map features
/***************************************************************************************************/
void compute_DD(InputArray image, int lowThreshold, int ratio, int r, OutputArray confidence_map)
{
	/*distance to edge - distance to border*/

	Mat _image_gray, _image, _edges, _dst;;

	//convert to CV_8U
	image.getMat().convertTo(_image, CV_8U);
	_dst.create(_image.size(), _image.depth());

	//convert the image to grayscale
	if (_image.channels() > 1)
		cvtColor(_image, _image_gray, CV_BGR2GRAY);
	else
		_image_gray = _image.clone();

	//canny edge detector
	blur(_image_gray, _edges, Size(3, 3));
	Canny(_edges, _edges, lowThreshold, lowThreshold * ratio, 2 * r + 1);

	_dst.setTo(Scalar(0));
	_image.copyTo(_dst, _edges);

	//set all countours to 0 (useful for distanceTransform)
	threshold(_dst, _dst, 0, 255, 1);

	//compute distance to edges
	distanceTransform(_dst, confidence_map, CV_DIST_L2, 3);
}

void disparity_variance_measure(InputArray disparity_map, int size, OutputArray confidence_map)
{
	confidence_map.create(disparity_map.size(), CV_32F);

	Mat _disparity_map = disparity_map.getMat(),
		_confidence_map = confidence_map.getMat(),
		_confidence_measure_x = Mat(disparity_map.size(), CV_32F),
		_confidence_measure_y = Mat(disparity_map.size(), CV_32F),
		_abs_confidence_measure_x,
		_abs_confidence_measure_y;

	Sobel(disparity_map, _confidence_measure_x, disparity_map.depth(), 1, 0, size, 1, 0, BORDER_DEFAULT);
	convertScaleAbs(_confidence_measure_x, _abs_confidence_measure_x);

	Sobel(disparity_map, _confidence_measure_y, disparity_map.depth(), 0, 1, size, 1, 0, BORDER_DEFAULT);
	convertScaleAbs(_confidence_measure_y, _abs_confidence_measure_y);

	for (int row = 0; row < _disparity_map.rows; row++)
	{
		uchar* abs_confidence_measure_x_ptr = _abs_confidence_measure_x.ptr<uchar>(row);
		uchar* abs_confidence_measure_y_ptr = _abs_confidence_measure_y.ptr<uchar>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < _disparity_map.cols; col++)
		{
			uchar x_value = abs_confidence_measure_x_ptr[col];
			uchar y_value = abs_confidence_measure_y_ptr[col];
			
			confidence_map_ptr[col] = - sqrt(pow(x_value, 2) + pow(y_value, 2));
		}
	}
}

void var_measure(InputArray disparity_map, int r, OutputArray confidence_map)
{
	int win = 2 * r + 1;

	confidence_map.create(disparity_map.size(), CV_32F);

	Mat _disparity_map = disparity_map.getMat(),
		_confidence_map = confidence_map.getMat(),
		_mean, _pow, _diff, _var;

	blur(_disparity_map, _mean, Size(win, win));
	_diff = _disparity_map - _mean;
	multiply(_diff, _diff, _pow);
	blur(_pow, _var, Size(win, win));
	_confidence_map = -_var;
}

void difference_with_median(InputArray disparity_L2R, int r,int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _disparity_L2R = Mat(height, width, CV_8U),
		_confidence_map = confidence_map.getMat(),
		_median;

	disparity_L2R.getMat().convertTo(_disparity_L2R, CV_8U);

	//compute median
	medianBlur(_disparity_L2R, _median, 2 * r + 1);

	for (int row = 0; row < height; row++)
	{
		uchar* _disparity_L2R_ptr = _disparity_L2R.ptr<uchar>(row);
		uchar* median_ptr = _median.ptr<uchar>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			confidence_map_ptr[col] = (float)-fabs(_disparity_L2R_ptr[col] - median_ptr[col]);
		}
	}
}

/***************************************************************************************************/
//	3.5 Reference image features
/***************************************************************************************************/
void distance_to_border(InputArray image, OutputArray confidence_map)
{
	Mat _image = image.getMat();

	int height = _image.rows;
	int width = _image.cols;

	Mat _border_map = Mat(height, width, CV_8U);
	_border_map.setTo(Scalar(1));

	confidence_map.create(_image.size(), CV_32F);
	Mat _confidence_map = confidence_map.getMat();
	
	for (int row = 0; row < height; row++)
	{
		uchar* border_map_ptr = _border_map.ptr<uchar>(row);

		for (int col = 0; col < width; col++)
		{
			if (row == 0) border_map_ptr[col] = 0;
			if (col == 0) border_map_ptr[col] = 0;
			if (row == height - 1) border_map_ptr[col] = 0;
			if (col == width  - 1) border_map_ptr[col] = 0;
		}
	}

	distanceTransform(_border_map, _confidence_map, CV_DIST_L2, 3);
}

void distance_to_left_border(InputArray image, int d_max, OutputArray confidence_map)
{
	Mat _image = image.getMat();

	int height = _image.rows;
	int width = _image.cols;

	Mat _border_map = Mat(height, width, CV_8U);
	_border_map.setTo(Scalar(1));

	confidence_map.create(_image.size(), CV_32F);
	Mat _confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		uchar* border_map_ptr = _border_map.ptr<uchar>(row);

		for (int col = 0; col < width; col++)
		{
			if (col == 0) border_map_ptr[col] = 0;
		}
	}

	distanceTransform(_border_map, _confidence_map, CV_DIST_L2, 3);
	threshold( _confidence_map, _confidence_map, d_max, d_max, THRESH_TRUNC);
	
}

void horizontal_gradient(InputArray left_stereo, OutputArray confidence_map)
{
	confidence_map.create(left_stereo.size(), CV_32F);
	Mat _confidence_map = confidence_map.getMat(),
        _left_stereo_gray, 
		_grad_x, 
		_grad_x_abs;

	//convert the image to grayscale
	if (left_stereo.getMat().channels() > 1)
		cvtColor(left_stereo, _left_stereo_gray, CV_BGR2GRAY);
	else
		_left_stereo_gray = left_stereo.getMat().clone();
	
	Sobel(_left_stereo_gray, _grad_x, left_stereo.depth(), 1, 0, 1, 1, 0, BORDER_DEFAULT);
	convertScaleAbs(_grad_x, _grad_x_abs);
	_grad_x_abs.convertTo(_confidence_map, CV_32F);
}

/***************************************************************************************************/
//	3.6 Image distinctiveness
/***************************************************************************************************/
void distinctiveness(InputArrayOfArrays costs_LL, int d_min, int d_max, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	vector<Mat> _costs_LL;
	costs_LL.getMatVector(_costs_LL);

	Mat min = confidence_map.getMat();

	min.setTo(Scalar(-1));

	float minimum;

	for (int row = 0; row < height; row++)
	{
		float* min_ptr = min.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			minimum = FLT_MAX;

			for (int d = 0; d < _costs_LL.size(); d++)
			{
				float value = _costs_LL[d].ptr<float>(row)[col];

				if (value < minimum && value >= 0 && d != - d_min)
				{
					minimum = value;
				}
			}

			min_ptr[col] = minimum;
		}
	}
}

void distinctive_similarity_measure(InputArray distinctiveness_L, InputArray distinctiveness_R, InputArray c1, InputArray c1_idx, float epsilon, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _distinctiveness_L = distinctiveness_L.getMat(),
		_distinctiveness_R = distinctiveness_R.getMat();

	Mat _c1 = c1.getMat(), 
        _c1_idx = c1_idx.getMat(),
        _confidence_map = confidence_map.getMat();

	for (int row = 0; row < height; row++)
	{
		float* c1_ptr = _c1.ptr<float>(row);
		float* distinctiveness_L_ptr = _distinctiveness_L.ptr<float>(row);
		float* distinctiveness_R_ptr = _distinctiveness_R.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		uchar* c1_idx_ptr = _c1_idx.ptr<uchar>(row);

		for (int col = 0; col < width; col++)
		{
			uchar d_1 = c1_idx_ptr[col]; 
			uchar index = col - d_1;

			if (index >= 0)
			{
				float d_l = distinctiveness_L_ptr[col];
				float d_r = distinctiveness_R_ptr[index];
				float c_1 = c1_ptr[col];

				confidence_map_ptr[col] = ((d_l * d_r) + epsilon) / ((c_1 * c_1) + epsilon);
			}
		}
	}
}

void self_aware_matching(InputArrayOfArrays costs, InputArrayOfArrays costs_LL, InputArray c1_idx, int d_min, int d_max, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _c1_idx = c1_idx.getMat(),
	    _confidence_map = confidence_map.getMat();

	vector<Mat> _costs; costs.getMatVector(_costs);
	vector<Mat> _costs_LL; costs_LL.getMatVector(_costs_LL);

	float sum, 
		  sum_mean_LR, sum_mean_LL, 
		  sum_deviation_LR, sum_deviation_LL,
          value_LR, value_LL,
		  mean_LR, mean_LL,
		  standard_deviation_LR, standard_deviation_LL;

	int count_LR, count_LL, 
		num_disp = d_max - d_min + 1;

	for (int row = 0; row < height; row++)
	{
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		uchar* _c1_idx_ptr = _c1_idx.ptr<uchar>(row);

		for (int col = 0; col < width; col++)
		{
			count_LR = 0, count_LL = 0;
			sum_mean_LR = 0, sum_mean_LL = 0;
			uchar d1 = _c1_idx_ptr[col]; 
			int offset = - d_min - d1; 

			//mean computation
			for (int d = 0; d < _costs.size(); d++)
			{
				if (d + offset < _costs_LL.size() && d + offset >= 0)
				{
					value_LR = _costs[d].ptr<float>(row)[col];
					sum_mean_LR += value_LR;
					count_LR++;

					value_LL = _costs_LL[d + offset].ptr<float>(row)[col];
					sum_mean_LL += value_LL;
					count_LL++;
				}
			}

			mean_LR = sum_mean_LR / count_LR;
			mean_LL = sum_mean_LL / count_LL;

			//standard deviation and SAMM computation
			sum = 0, sum_deviation_LR = 0, sum_deviation_LL = 0;

			for (int d = 0; d < _costs.size(); d++)
			{
				if (d + offset < _costs_LL.size() && d + offset >= 0)
				{
					value_LR = _costs[d].ptr<float>(row)[col];
					sum_deviation_LR += pow(value_LR - mean_LR, 2);

					value_LL = _costs_LL[d + offset].ptr<float>(row)[col];
					sum_deviation_LL += pow(value_LL - mean_LL, 2);

					sum += (value_LR - mean_LR)*(value_LL - mean_LL);
				}
			}

			standard_deviation_LR = sqrt(sum_deviation_LR);
			standard_deviation_LL = sqrt(sum_deviation_LL);

			confidence_map_ptr[col] = sum / sqrt(sum_deviation_LR * sum_deviation_LL);
		}
	}
}

/***************************************************************************************************/
//	3.8 SGM specific
/***************************************************************************************************/
void PS_measure(Mat cl1, Mat cl2, Mat cl1_idx, Mat cl2_idx, Mat cg1_idx, int height, int width, float gamma, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);
	Mat PS = confidence_map.getMat();

	float first, second, third, minTerm;
	for (int row = 0; row < height; row++)
		for (int col = 0; col < width; col++)
		{
			first = (cl2.at<float>(row,col) - cl1.at<float>(row,col))/(cl1.at<float>(row,col)+0.1);

			minTerm = abs(cl1_idx.at<uchar>(row,col) - cl2_idx.at<uchar>(row,col)) < gamma ? abs(cl1_idx.at<uchar>(row,col) - cl2_idx.at<uchar>(row,col)) : gamma;
			second = 1 - minTerm / gamma;

			minTerm = abs(cl1_idx.at<uchar>(row,col) - cg1_idx.at<uchar>(row,col)) < gamma ? abs(cl1_idx.at<uchar>(row,col) - cg1_idx.at<uchar>(row,col)) : gamma;
			third = 1 - minTerm / gamma; 

			// PS
			PS.at<float>(row, col) = first * second * third;
			//printf("%f = %f * %f * %f\n", PS.at<float>(row, col), first, second, third);
		}					
}

void sum_of_consistent_disparites(InputArray disparity_map, InputArrayOfArrays scanline_disparity_maps, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _disparity_map = disparity_map.getMat(),
		_confidence_map = confidence_map.getMat();

	vector<Mat> _scanline_disparity_maps; 
	scanline_disparity_maps.getMatVector(_scanline_disparity_maps);

	for (int row = 0; row < height; row++)
	{
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);
		float* disparity_map_ptr = _disparity_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			int sum = 0;

			for (int scanline = 0; scanline < _scanline_disparity_maps.size(); scanline++)
			{
				float scanline_disparity_value = _scanline_disparity_maps.at(scanline).ptr<float>(row)[col];
				float disparity_value = disparity_map_ptr[col];

				if (disparity_value == scanline_disparity_value) sum++;
			}

			confidence_map_ptr[col] = sum; 
		}
	}
}

/***************************************************************************************************/
//	More functions:
/***************************************************************************************************/

//	Compute the error to ground-truth
void error_to_groundtruth(InputArray disparity_L2R, InputArray groundtruth, int height, int width, OutputArray confidence_map)
{
	confidence_map.create(height, width, CV_32F);

	Mat _disparity_L2R = Mat(height, width, CV_8U),
		_groundtruth = groundtruth.getMat(),
		_confidence_map = confidence_map.getMat();

	disparity_L2R.getMat().convertTo(_disparity_L2R, CV_8U);

	for (int row = 0; row < height; row++)
	{
		uchar* _disparity_L2R_ptr = _disparity_L2R.ptr<uchar>(row);
		float* groundtruth_ptr = _groundtruth.ptr<float>(row);
		float* confidence_map_ptr = _confidence_map.ptr<float>(row);

		for (int col = 0; col < width; col++)
		{
			confidence_map_ptr[col] = (float)(_disparity_L2R_ptr[col] - groundtruth_ptr[col]);
		}
	}
}

//	Sample functions processing (most of the) confidence measures
void fn_confidence_measure(InputArray image_L, InputArray image_R, _DSI dsi_LR, _DSI dsi_LL, _DSI dsi_RR, confParams params, OutputArray disparity_map, vector<string> &methods, OutputArrayOfArrays confidences)
{
	//computation
	printf("\n\n ***** Compute Base Costs *****\n\n");

	//confidence measures 
	vector<Mat> _confidences;

	//base info
	int height = dsi_LR.height;
	int width = dsi_LR.width;
	int d_min = dsi_LR.d_min;
	int d_max = dsi_LR.d_max;
	int num_disp = dsi_LR.num_disp;
	int disp_scale = 256 / num_disp;
	vector<Mat> costs = dsi_LR.values;
	vector<Mat> costs_LL = dsi_LL.values;
	vector<Mat> costs_RR = dsi_RR.values;

	//generate right dsi.
	printf(" - generate right dsi...\n");
	_DSI dsi_RL = DSI_left2right(dsi_LR);

	//compute left and right disparity maps
	printf(" - compute left and right disparity maps...\n");
	Mat disparity_L2R = disparity_map_L2R(dsi_LR);
	Mat disparity_R2L = disparity_map_R2L(dsi_LR);

	//compute c_1, c_2, c^_2 in the paper as well as sum of matching costs and
	//number of inflection points(NOI).
	printf(" - compute c_1, c_2, c^_2 in the paper as well as sum of matching costs...\n");
	vector<Mat> local_minima;
	Mat c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, c1_R; 
	compute_base_costs(dsi_LR.values, c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, local_minima);
	clock_t start, end;
	float seconds;

	//compute minimum right to left
	printf(" - compute minimum right to left...\n");
	minimum(dsi_RL.values, c1_R);

	//computation
	printf("\n\n ***** Confidence Measure  *****\n\n");

	// 01. Matching Score Measure (MSM).
	if (params.useMSM == true)
	{
		Mat msm;

		printf(" - confidence measure: matching score measure (MSM)\n");
		matching_score_measure(c1, height, width, msm);



		_confidences.push_back(msm);
		methods.push_back("msm");
	}

	// 02. Local Curve (LC).
	if (params.useLC == true)
	{
		Mat lc;

		printf(" - confidence measure: local curve (LC)\n");
		local_curve(costs, c1_idx, params.lc_gamma, height, width, lc);



		_confidences.push_back(lc);
		methods.push_back("lc");
	}

	// 03. Curvature (CUR).
	if (params.useCUR == true)
	{
		Mat cur;

		printf(" - confidence measure: curvature (CUR)\n");
		curvature(costs, c1_idx, height, width, cur);



		_confidences.push_back(cur);
		methods.push_back("cur");
	}

	// 04. Peak Ratio (PKR).
	if (params.usePKR == true)
	{
		Mat pkr;

		printf(" - confidence measure: peak ratio (PKR)\n");
		peak_ratio(c1, c_hat_2, params.pkr_epsilon, height, width, pkr);



		_confidences.push_back(pkr);
		methods.push_back("pkr");
	}

	// 05. Peak Ratio Naive (PKRN).
	if (params.usePKRN == true)
	{
		Mat pkrn;

		printf(" - confidence measure: peak ratio naive (PKRN)\n");
		peak_ratio(c1, c2, params.pkr_epsilon, height, width, pkrn);



		_confidences.push_back(pkrn);
		methods.push_back("pkrn");
	}

	// 06. Average Peak Ratio (APKR).
	if (params.useAPKR == true)
	{
		Mat apkr5, apkr7, apkr9, apkr11;

		printf(" - confidence measure: average peak ratio (APKR)\n");
		average_peak_ratio(costs, c1_idx, c_hat_2_idx, params.apkr_epsilon, 2, height, width, apkr5);



		_confidences.push_back(apkr5);
		methods.push_back("apkr_5");

		printf(" - confidence measure: average peak ratio (APKR)\n");
		average_peak_ratio(costs, c1_idx, c_hat_2_idx, params.apkr_epsilon, 3, height, width, apkr7);



		_confidences.push_back(apkr7);
		methods.push_back("apkr_7");

		printf(" - confidence measure: average peak ratio (APKR)\n");
		average_peak_ratio(costs, c1_idx, c_hat_2_idx, params.apkr_epsilon, 4, height, width, apkr9);



		_confidences.push_back(apkr9);
		methods.push_back("apkr_9");

		printf(" - confidence measure: average peak ratio (APKR)\n");
		average_peak_ratio(costs, c1_idx, c_hat_2_idx, params.apkr_epsilon, 5, height, width, apkr11);



		_confidences.push_back(apkr11);
		methods.push_back("apkr_11");
	}

	// 07. Average Peak Ratio Naive (APKRN).
	if (params.useAPKRN == true)
	{
		Mat apkrn5;

		printf(" - confidence measure: average peak ratio naive (APKRN)\n");
		average_peak_ratio(costs, c1_idx, c2_idx, params.apkr_epsilon, 2, height, width, apkrn5);



		_confidences.push_back(apkrn5);
		methods.push_back("apkrn_5");

		Mat apkrn7;

		printf(" - confidence measure: average peak ratio naive (APKRN)\n");
		average_peak_ratio(costs, c1_idx, c2_idx, params.apkr_epsilon, 3, height, width, apkrn7);



		_confidences.push_back(apkrn7);
		methods.push_back("apkrn_7");

		Mat apkrn9;

		printf(" - confidence measure: average peak ratio naive (APKRN)\n");
		average_peak_ratio(costs, c1_idx, c2_idx, params.apkr_epsilon, 4, height, width, apkrn9);



		_confidences.push_back(apkrn9);
		methods.push_back("apkrn_9");

		Mat apkrn11;

		printf(" - confidence measure: average peak ratio naive (APKRN)\n");
		average_peak_ratio(costs, c1_idx, c2_idx, params.apkr_epsilon, 5, height, width, apkrn11);



		_confidences.push_back(apkrn11);
		methods.push_back("apkrn_11");

	}

	// 08. Maximum Margin (MM).
	if (params.useMM == true)
	{
		Mat mm;

		printf(" - confidence measure: maximum margin (MM)\n");
		maximum_margin(c1, c_hat_2, height, width, mm);



		_confidences.push_back(mm);
		methods.push_back("mm");
	}

	// 09. Maximum Margin Naive (MMN).
	if (params.useMMN == true)
	{
		Mat mmn;

		printf(" - confidence measure: maximum margin naive (MMN)\n");
		maximum_margin(c1, c2, height, width, mmn);



		_confidences.push_back(mmn);
		methods.push_back("mmn");
	}

	// 10. Nonlinear Maximum Margin
	if (params.useNMM == true)
	{
		Mat nmm;

		printf(" - confidence measure: nonlinear margin (NMM)\n");
		non_linear_margin(c1, c_hat_2, params.nmm_sigma, height, width, nmm);



		_confidences.push_back(nmm);
		methods.push_back("nmm");
	}

	// 11. Nonlinear Maximum Margin Naive
	if (params.useNMMN == true)
	{
		Mat nmmn;

		printf(" - confidence measure: nonlinear margin naive (NMMN)\n");
		non_linear_margin(c1, c2, params.nmm_sigma, height, width, nmmn);



		_confidences.push_back(nmmn);
		methods.push_back("nmmn");
	}

	// 12. Disparity Variance Measure (DVM).
	if (params.useDVM == true)
	{
		Mat dvm3;

		printf(" - confidence measure: disparity variance measure (DVM)\n");

		disparity_variance_measure(disparity_L2R, 3, dvm3);



		_confidences.push_back(dvm3);
		methods.push_back("dvm");
	}

	// 13. Disparity Ambiguity Measure (DAM).
	if (params.useDAM == true)
	{
		Mat dam;

		printf(" - confidence measure: disparity ambiguity measure (DAM)\n");
		disparity_ambiguity_measure(c1_idx, c2_idx, height, width, dam);



		_confidences.push_back(dam);
		methods.push_back("dam");
	}

	// 14. Maximum Likelihood Measure (MLM).
	if (params.useMLM == true)
	{
		Mat mlm;

		printf(" - confidence measure: maximum likelihood measure (MLM)\n");
		maximum_likelihood_measure(costs, c1, params.mlm_sigma, height, width, mlm);



		_confidences.push_back(mlm);
		methods.push_back("mlm");
	}

	// 15. Attainable Maximum Likelihood (AML).
	if (params.useAML == true)
	{
		Mat aml;

		printf(" - confidence measure: attainable maximum likelihood (AML)\n");
		attainable_maximum_likelihood(costs, c1, params.aml_sigma, height, width, aml);



		_confidences.push_back(aml);
		methods.push_back("aml");
	}

	// 16. Negative Entropy  (NEM).
	if (params.useNEM == true)
	{
		Mat nem;

		printf(" - confidence measure: negative entropy (NEM)\n");
		negative_entropy_measure(costs, height, width, nem);



		_confidences.push_back(nem);
		methods.push_back("nem");
	}

	// 17. Number Of Inflections (NOI).
	if (params.useNOI == true)
	{

		printf(" - confidence measure: number of inflections (NOI)\n");



		_confidences.push_back(NOI);
		methods.push_back("noi");
	}

	// 18. Winner Margin (WMN).
	if (params.useWMN == true)
	{
		Mat wmn;

		printf(" - confidence measure: winner margin (WMN)\n");
		winner_margin(c1, c_hat_2, c_sum, height, width, wmn);



		_confidences.push_back(wmn);
		methods.push_back("wmn");
	}

	// 19. Winner Margin Naive (WMNN).
	if (params.useWMNN == true)
	{
		Mat wmnn;

		printf(" - confidence measure: winner margin naive (WMNN)\n");
		winner_margin(c1, c2, c_sum, height, width, wmnn);



		_confidences.push_back(wmnn);
		methods.push_back("wmnn");
	}

	// 20. Perturbation Measure (PER).
	if (params.usePER == true)
	{
		Mat per;

		printf(" - confidence measure: perturbation measure (PER)\n");
		perturbation_measure(costs, c1, c1_idx, params.per_sigma, height, width, per);



		_confidences.push_back(per);
		methods.push_back("per");
	}

	// 21. Left Right Consistency Check (LRC).
	if (params.useLRC == true)
	{
		Mat lrc;

		printf(" - confidence measure: left right consistency check (LRC)\n");
		left_right_consistency_check(disparity_L2R, disparity_R2L, disp_scale, height, width, lrc);



		_confidences.push_back(lrc);
		methods.push_back("lrc");
	}

	// 22. Asymmetric Consistency Check (ACC).
	if (params.useACC == true)
	{
		Mat acc;

		printf(" - confidence measure: asymmetric consistency check (ACC)\n");
		asymmetric_consistency_check(c1, c1_idx, height, width, acc);



		_confidences.push_back(acc);
		methods.push_back("acc");
	}

	// 23. Uniqueness Constraint (UC).
	if (params.useUC == true)
	{
		Mat uc_binary, uc_occurence, uc_cost;

		printf(" - confidence measure: uniqueness constraint (UC)\n");
		uniqueness_constraint(c1, c1_idx, height, width, uc_binary, uc_occurence, uc_cost);



		_confidences.push_back(uc_binary);



		_confidences.push_back(uc_occurence);



		_confidences.push_back(uc_cost);
		methods.push_back("uc_binary");
		methods.push_back("uc_occurence");
		methods.push_back("uc_cost");
	}

	// 24. Left Right Difference (LRD).
	if (params.useLRD == true)
	{
		Mat lrd;

		printf(" - confidence measure: left right difference (LRD)\n");
		left_right_difference(c1, c2, c1_R, disparity_L2R, disp_scale, params.lrd_epsilon, height, width, lrd);



		_confidences.push_back(lrd);
		methods.push_back("lrd");
	}

	// 25. Horizontal Gradient (HGM).
	if (params.useHGM == true)
	{
		Mat hgm;

		printf(" - confidence measure: horizontal gradient (HGM)\n");
		horizontal_gradient(image_L, hgm);



		_confidences.push_back(hgm);
		methods.push_back("hgm");
	}

	// 26. Zero Mean Sum Of Absolute Difference (ZSAD).
	if (params.useZSAD == true)
	{
		Mat zsad;

		printf(" - confidence measure: zero mean sum of absolute difference (ZSAD)\n");
		zero_mean_sum_of_absolute_differences(image_L, image_R, c1, c1_idx, params.zsad_radius, height, width, zsad);



		_confidences.push_back(zsad);
		methods.push_back("zsad");
	}

	// 27. Distinctiveness (DTS).
	if (params.useDTS == true)
	{
		Mat dts_L;

		printf(" - confidence measure: distinctiveness (DTS)\n");
		distinctiveness(costs_LL, dsi_LL.d_min, dsi_LL.d_max, height, width, dts_L);



		_confidences.push_back(dts_L);
		methods.push_back("dts");
	}

	// 28. Distinctive Similarity Measure (DSM).
	if (params.useDSM == true)
	{
		Mat dsm, distinctiveness_R, distinctiveness_L;

		printf(" - confidence measure: distinctive similarity measure (DSM)\n");

		distinctiveness(costs_LL, dsi_LL.d_min, dsi_LL.d_max, height, width, distinctiveness_L);
		distinctiveness(costs_RR, dsi_RR.d_min, dsi_RR.d_max, height, width, distinctiveness_R);

		distinctive_similarity_measure(distinctiveness_L, distinctiveness_R, c1, c1_idx, params.dsm_epsilon, height, width, dsm);



		_confidences.push_back(dsm);
		methods.push_back("dsm");
	}

	// 29. Self Aware Matching Measure(SAMM).
	if (params.useSAMM == true)
	{
		Mat samm;

		printf(" - confidence measure: self aware matching measure (SAMM)\n");
		self_aware_matching(dsi_LR.values, dsi_LL.values, c1_idx, dsi_LL.d_min, dsi_LL.d_max, height, width, samm);



		_confidences.push_back(samm);
		methods.push_back("samm");
	}

	// 30. Distance To Border (DB).
	if (params.useDB == true)
	{
		Mat db;

		printf(" - confidence measure: distance to border (DB)\n");
		distance_to_border(disparity_L2R, db);



		_confidences.push_back(db);
		methods.push_back("db");
	}

	// 31. Distance To Border Left (DBL).
	if (params.useDBL == true)
	{
		Mat dbl;

		printf(" - confidence measure: distance to left border (DBL)\n");
		distance_to_left_border(disparity_L2R, dsi_LL.d_max, dbl);



		_confidences.push_back(dbl);
		methods.push_back("dbl");
	}

	// 32. Distance To Discontinuity (DTD).
	if (params.useDTD == true)
	{
		Mat dtd;

		printf(" - confidence measure: distance to discontinuity (DTD)\n");
		compute_DD(disparity_L2R, params.DD_max_lowThreshold, params.DD_ratio, params.DD_radius, dtd);



		_confidences.push_back(dtd);
		methods.push_back("dtd");
	}

	// 33. Distance To Edge (DTE).
	if (params.useDTE == true)
	{
		Mat dte;

		printf(" - confidence measure: distance to edge (DTE)\n");
		compute_DD(image_L, params.DD_max_lowThreshold, params.DD_ratio, params.DD_radius, dte);



		_confidences.push_back(dte);
		methods.push_back("dte");
	}

	// 34. Difference With Median (MED).
	if (params.useMED == true)
	{
		Mat med5, med7, med9, med11;

		printf(" - confidence measure: difference with median (MED)\n");

		difference_with_median(disparity_L2R, 2, height, width, med5);



		_confidences.push_back(med5);
		methods.push_back("med5");

		difference_with_median(disparity_L2R, 3, height, width, med7);



		_confidences.push_back(med7);
		methods.push_back("med7");

		difference_with_median(disparity_L2R, 4, height, width, med9);



		_confidences.push_back(med9);
		methods.push_back("med9");

		difference_with_median(disparity_L2R, 5, height, width, med11);



		_confidences.push_back(med11);
		methods.push_back("med11");

	}

	// 34. Difference With Median (MED).
	if (params.useVAR == true)
	{
		Mat var5, var7, var9, var11;

		printf(" - confidence measure: variance (VAR)\n");

		var_measure(disparity_L2R, 2, var5);



		_confidences.push_back(var5);
		methods.push_back("var5");

		var_measure(disparity_L2R, 3, var7);



		_confidences.push_back(var7);
		methods.push_back("var7");

		var_measure(disparity_L2R, 4, var9);



		_confidences.push_back(var9);
		methods.push_back("var9");

		var_measure(disparity_L2R, 5, var11);



		_confidences.push_back(var11);
		methods.push_back("var11");

	}

	// 35. Local Minima In Neighborhood (LMN).
	if (params.useLMN == true)
	{
		Mat lmn;

		printf(" - confidence measure: local minima in neighborhood (LMN)\n");
		local_minima_in_neighborhood(local_minima, c1_idx, params.lmn_radius, height, width, lmn);



		_confidences.push_back(lmn);
		methods.push_back("lmn");
	}

	// 36. Weighted Peak Ratio (WPKR).
	if (params.useWPKR == true)
	{
		Mat wpkr5;

		printf(" - confidence measure: weighted peak ratio (WPKR)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c_hat_2_idx, params.wpkr_epsilon, params.wpkr_threshold, 2, height, width, wpkr5);



		_confidences.push_back(wpkr5);
		methods.push_back("wpkr_5");

		Mat wpkr7;

		printf(" - confidence measure: weighted peak ratio (WPKR)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c_hat_2_idx, params.wpkr_epsilon, params.wpkr_threshold, 3, height, width, wpkr7);



		_confidences.push_back(wpkr7);
		methods.push_back("wpkr_7");

		Mat wpkr9;

		printf(" - confidence measure: weighted peak ratio (WPKR)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c_hat_2_idx, params.wpkr_epsilon, params.wpkr_threshold, 4, height, width, wpkr9);



		_confidences.push_back(wpkr9);
		methods.push_back("wpkr_9");

		Mat wpkr11;

		printf(" - confidence measure: weighted peak ratio (WPKR)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c_hat_2_idx, params.wpkr_epsilon, params.wpkr_threshold, 5, height, width, wpkr11);



		_confidences.push_back(wpkr11);
		methods.push_back("wpkr_11");

	}

	// 37. Weighted Peak Ratio Naive (WPKRN).
	if (params.useWPKRN == true)
	{
		Mat wpkrn5;

		printf(" - confidence measure: weighted peak ratio naive (WPKRN)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c2_idx, params.wpkr_epsilon, params.wpkr_threshold, 2, height, width, wpkrn5);



		_confidences.push_back(wpkrn5);
		methods.push_back("wpkrn_5");

		Mat wpkrn7;

		printf(" - confidence measure: weighted peak ratio naive (WPKRN)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c2_idx, params.wpkr_epsilon, params.wpkr_threshold, 3, height, width, wpkrn7);



		_confidences.push_back(wpkrn7);
		methods.push_back("wpkrn_7");

		Mat wpkrn9;

		printf(" - confidence measure: weighted peak ratio naive (WPKRN)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c2_idx, params.wpkr_epsilon, params.wpkr_threshold, 4, height, width, wpkrn9);



		_confidences.push_back(wpkrn9);
		methods.push_back("wpkrn_9");

		Mat wpkrn11;

		printf(" - confidence measure: weighted peak ratio naive (WPKRN)\n");
		weighted_peak_ratio(costs, image_L, c1_idx, c2_idx, params.wpkr_epsilon, params.wpkr_threshold, 5, height, width, wpkrn11);



		_confidences.push_back(wpkrn11);
		methods.push_back("wpkrn_11");
	}


	//confidence maps
	confidences.create(_confidences.size(), 1, Mat(_confidences).type());

	for (int i = 0; i < _confidences.size(); i++)
		confidences.getMatRef(i) = _confidences[i];
	
	//disparity map 
	disparity_map.getMatRef() = disparity_L2R;
}

